@@ -1,10 +1,24 @@
 use ref_cast::RefCast;
 use serde::{Deserialize, Serialize};
+use std::borrow::Cow;
 use std::fmt::{Debug, Display};
 use std::ops::{Deref, DerefMut};
 use std::str::FromStr;
+use thiserror::Error;
 use url::Url;
 
+#[derive(Error, Debug, Clone, PartialEq, Eq)]
+pub enum DisplaySafeUrlError {
+    /// Failed to parse a URL.
+    #[error(transparent)]
+    Url(#[from] url::ParseError),
+
+    /// We parsed a URL, but couldn't disambiguate its authority
+    /// component.
+    #[error("ambiguous user/pass authority in URL (not percent-encoded?): {0}")]
+    AmbiguousAuthority(String),
+}
+
 /// A [`Url`] wrapper that redacts credentials when displaying the URL.
 ///
 /// `DisplaySafeUrl` wraps the standard [`url::Url`] type, providing functionality to mask
@@ -46,8 +60,55 @@ pub struct DisplaySafeUrl(Url);
 
 impl DisplaySafeUrl {
     #[inline]
-    pub fn parse(input: &str) -> Result<Self, url::ParseError> {
-        Ok(Self(Url::parse(input)?))
+    pub fn parse(input: &str) -> Result<Self, DisplaySafeUrlError> {
+        let url = Url::parse(input)?;
+
+        // Reject some ambiguous cases, e.g., `https://user/name:password@domain/a/b/c`
+        //
+        // In this case the user *probably* meant to have a username of "user/name", but both RFC
+        // 3986 and WHATWG URL expect the userinfo (RFC 3986) or authority (WHATWG) to not contain a
+        // non-percent-encoded slash or other special character.
+        //
+        // This ends up being moderately annoying to detect, since the above gets parsed into a
+        // "valid" WHATWG URL where the host is `used` and the pathname is
+        // `/name:password@domain/a/b/c` rather than causing a parse error.
+        //
+        // To detect it, we use a heuristic: if the password component is missing but the path or
+        // fragment contain a `:` followed by a `@`, then we assume the URL is ambiguous.
+        if url.password().is_none()
+            && (url
+                .path()
+                .find(':')
+                .is_some_and(|pos| url.path()[pos..].contains('@'))
+                || url
+                    .fragment()
+                    .map(|fragment| {
+                        fragment
+                            .find(':')
+                            .is_some_and(|pos| fragment[pos..].contains('@'))
+                    })
+                    .unwrap_or(false))
+            // If the above is true, we should always expect to find these in the given URL
+            && let Some(col_pos) = input.find(':')
+            && let Some(at_pos) = input.rfind('@')
+        {
+            // Our ambiguous URL probably has credentials in it, so we don't want to blast it out in
+            // the error message. We somewhat aggressively replace everything between the scheme's
+            // ':' and the lastmost `@` with `***`.
+            let redacted_path = format!("{}***{}", &input[0..=col_pos], &input[at_pos..]);
+            return Err(DisplaySafeUrlError::AmbiguousAuthority(redacted_path));
+        }
+
+        Ok(Self(url))
+    }
+
+    /// Create a new [`DisplaySafeUrl`] from a [`Url`].
+    ///
+    /// Unlike [`Self::parse`], this doesn't perform any ambiguity checks.
+    /// That means that it's primarily useful for contexts where a human can't easily accidentally
+    /// introduce an ambiguous URL, such as URLs being read from a request.
+    pub fn from_url(url: Url) -> Self {
+        Self(url)
     }
 
     /// Cast a `&Url` to a `&DisplaySafeUrl` using ref-cast.
@@ -58,8 +119,8 @@ impl DisplaySafeUrl {
 
     /// Parse a string as an URL, with this URL as the base URL.
     #[inline]
-    pub fn join(&self, input: &str) -> Result<Self, url::ParseError> {
-        self.0.join(input).map(DisplaySafeUrl::from)
+    pub fn join(&self, input: &str) -> Result<Self, DisplaySafeUrlError> {
+        Ok(Self(self.0.join(input)?))
     }
 
     /// Serialize with Serde using the internal representation of the `Url` struct.
@@ -77,12 +138,12 @@ impl DisplaySafeUrl {
     where
         D: serde::Deserializer<'de>,
     {
-        Url::deserialize_internal(deserializer).map(DisplaySafeUrl::from)
+        Ok(Self(Url::deserialize_internal(deserializer)?))
     }
 
     #[allow(clippy::result_unit_err)]
-    pub fn from_file_path<P: AsRef<std::path::Path>>(path: P) -> Result<DisplaySafeUrl, ()> {
-        Url::from_file_path(path).map(DisplaySafeUrl::from)
+    pub fn from_file_path<P: AsRef<std::path::Path>>(path: P) -> Result<Self, ()> {
+        Ok(Self(Url::from_file_path(path)?))
     }
 
     /// Remove the credentials from a URL, allowing the generic `git` username (without a password)
@@ -91,13 +152,31 @@ impl DisplaySafeUrl {
     pub fn remove_credentials(&mut self) {
         // For URLs that use the `git` convention (i.e., `ssh://git@github.com/...`), avoid dropping the
         // username.
-        if self.0.scheme() == "ssh" && self.0.username() == "git" && self.0.password().is_none() {
+        if is_ssh_git_username(&self.0) {
             return;
         }
         let _ = self.0.set_username("");
         let _ = self.0.set_password(None);
     }
 
+    /// Returns the URL with any credentials removed.
+    pub fn without_credentials(&self) -> Cow<'_, Url> {
+        if self.0.password().is_none() && self.0.username() == "" {
+            return Cow::Borrowed(&self.0);
+        }
+
+        // For URLs that use the `git` convention (i.e., `ssh://git@github.com/...`), avoid dropping the
+        // username.
+        if is_ssh_git_username(&self.0) {
+            return Cow::Borrowed(&self.0);
+        }
+
+        let mut url = self.0.clone();
+        let _ = url.set_username("");
+        let _ = url.set_password(None);
+        Cow::Owned(url)
+    }
+
     /// Returns [`Display`] implementation that doesn't mask credentials.
     #[inline]
     pub fn displayable_with_credentials(&self) -> impl Display {
@@ -128,7 +207,11 @@ impl Display for DisplaySafeUrl {
 impl Debug for DisplaySafeUrl {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
         let url = &self.0;
-        let (username, password) = if url.username() != "" && url.password().is_some() {
+        // For URLs that use the `git` convention (i.e., `ssh://git@github.com/...`), avoid masking the
+        // username.
+        let (username, password) = if is_ssh_git_username(url) {
+            (url.username(), None)
+        } else if url.username() != "" && url.password().is_some() {
             (url.username(), Some("****"))
         } else if url.username() != "" {
             ("****", None)
@@ -152,12 +235,6 @@ impl Debug for DisplaySafeUrl {
     }
 }
 
-impl From<Url> for DisplaySafeUrl {
-    fn from(url: Url) -> Self {
-        DisplaySafeUrl(url)
-    }
-}
-
 impl From<DisplaySafeUrl> for Url {
     fn from(url: DisplaySafeUrl) -> Self {
         url.0
@@ -165,13 +242,19 @@ impl From<DisplaySafeUrl> for Url {
 }
 
 impl FromStr for DisplaySafeUrl {
-    type Err = url::ParseError;
+    type Err = DisplaySafeUrlError;
 
     fn from_str(input: &str) -> Result<Self, Self::Err> {
-        Ok(Self(Url::from_str(input)?))
+        Self::parse(input)
     }
 }
 
+fn is_ssh_git_username(url: &Url) -> bool {
+    matches!(url.scheme(), "ssh" | "git+ssh" | "git+https")
+        && url.username() == "git"
+        && url.password().is_none()
+}
+
 fn display_with_redacted_credentials(
     url: &Url,
     f: &mut std::fmt::Formatter<'_>,
@@ -180,6 +263,12 @@ fn display_with_redacted_credentials(
         return write!(f, "{url}");
     }
 
+    // For URLs that use the `git` convention (i.e., `ssh://git@github.com/...`), avoid dropping the
+    // username.
+    if is_ssh_git_username(url) {
+        return write!(f, "{url}");
+    }
+
     write!(f, "{}://", url.scheme())?;
 
     if url.username() != "" && url.password().is_some() {
@@ -215,52 +304,65 @@ mod tests {
     #[test]
     fn from_url_no_credentials() {
         let url_str = "https://pypi-proxy.fly.dev/basic-auth/simple";
-        let url = Url::parse(url_str).unwrap();
-        let log_safe_url = DisplaySafeUrl::from(url);
+        let log_safe_url =
+            DisplaySafeUrl::parse("https://pypi-proxy.fly.dev/basic-auth/simple").unwrap();
         assert_eq!(log_safe_url.username(), "");
         assert!(log_safe_url.password().is_none());
-        assert_eq!(format!("{log_safe_url}"), url_str);
+        assert_eq!(log_safe_url.to_string(), url_str);
     }
 
     #[test]
     fn from_url_username_and_password() {
-        let url_str = "https://user:pass@pypi-proxy.fly.dev/basic-auth/simple";
-        let url = Url::parse(url_str).unwrap();
-        let log_safe_url = DisplaySafeUrl::from(url);
+        let log_safe_url =
+            DisplaySafeUrl::parse("https://user:pass@pypi-proxy.fly.dev/basic-auth/simple")
+                .unwrap();
         assert_eq!(log_safe_url.username(), "user");
         assert!(log_safe_url.password().is_some_and(|p| p == "pass"));
         assert_eq!(
-            format!("{log_safe_url}"),
+            log_safe_url.to_string(),
             "https://user:****@pypi-proxy.fly.dev/basic-auth/simple"
         );
     }
 
     #[test]
     fn from_url_just_password() {
-        let url_str = "https://:pass@pypi-proxy.fly.dev/basic-auth/simple";
-        let url = Url::parse(url_str).unwrap();
-        let log_safe_url = DisplaySafeUrl::from(url);
+        let log_safe_url =
+            DisplaySafeUrl::parse("https://:pass@pypi-proxy.fly.dev/basic-auth/simple").unwrap();
         assert_eq!(log_safe_url.username(), "");
         assert!(log_safe_url.password().is_some_and(|p| p == "pass"));
         assert_eq!(
-            format!("{log_safe_url}"),
+            log_safe_url.to_string(),
             "https://:****@pypi-proxy.fly.dev/basic-auth/simple"
         );
     }
 
     #[test]
     fn from_url_just_username() {
-        let url_str = "https://user@pypi-proxy.fly.dev/basic-auth/simple";
-        let url = Url::parse(url_str).unwrap();
-        let log_safe_url = DisplaySafeUrl::from(url);
+        let log_safe_url =
+            DisplaySafeUrl::parse("https://user@pypi-proxy.fly.dev/basic-auth/simple").unwrap();
         assert_eq!(log_safe_url.username(), "user");
         assert!(log_safe_url.password().is_none());
         assert_eq!(
-            format!("{log_safe_url}"),
+            log_safe_url.to_string(),
             "https://****@pypi-proxy.fly.dev/basic-auth/simple"
         );
     }
 
+    #[test]
+    fn from_url_git_username() {
+        let ssh_str = "ssh://git@github.com/org/repo";
+        let ssh_url = DisplaySafeUrl::parse(ssh_str).unwrap();
+        assert_eq!(ssh_url.username(), "git");
+        assert!(ssh_url.password().is_none());
+        assert_eq!(ssh_url.to_string(), ssh_str);
+        // Test again for the `git+ssh` scheme
+        let git_ssh_str = "git+ssh://git@github.com/org/repo";
+        let git_ssh_url = DisplaySafeUrl::parse(git_ssh_str).unwrap();
+        assert_eq!(git_ssh_url.username(), "git");
+        assert!(git_ssh_url.password().is_none());
+        assert_eq!(git_ssh_url.to_string(), git_ssh_str);
+    }
+
     #[test]
     fn parse_url_string() {
         let url_str = "https://user:pass@pypi-proxy.fly.dev/basic-auth/simple";
@@ -268,7 +370,7 @@ mod tests {
         assert_eq!(log_safe_url.username(), "user");
         assert!(log_safe_url.password().is_some_and(|p| p == "pass"));
         assert_eq!(
-            format!("{log_safe_url}"),
+            log_safe_url.to_string(),
             "https://user:****@pypi-proxy.fly.dev/basic-auth/simple"
         );
     }
@@ -281,17 +383,34 @@ mod tests {
         assert_eq!(log_safe_url.username(), "");
         assert!(log_safe_url.password().is_none());
         assert_eq!(
-            format!("{log_safe_url}"),
+            log_safe_url.to_string(),
             "https://pypi-proxy.fly.dev/basic-auth/simple"
         );
     }
 
+    #[test]
+    fn preserve_ssh_git_username_on_remove_credentials() {
+        let ssh_str = "ssh://git@pypi-proxy.fly.dev/basic-auth/simple";
+        let mut ssh_url = DisplaySafeUrl::parse(ssh_str).unwrap();
+        ssh_url.remove_credentials();
+        assert_eq!(ssh_url.username(), "git");
+        assert!(ssh_url.password().is_none());
+        assert_eq!(ssh_url.to_string(), ssh_str);
+        // Test again for `git+ssh` scheme
+        let git_ssh_str = "git+ssh://git@pypi-proxy.fly.dev/basic-auth/simple";
+        let mut git_shh_url = DisplaySafeUrl::parse(git_ssh_str).unwrap();
+        git_shh_url.remove_credentials();
+        assert_eq!(git_shh_url.username(), "git");
+        assert!(git_shh_url.password().is_none());
+        assert_eq!(git_shh_url.to_string(), git_ssh_str);
+    }
+
     #[test]
     fn displayable_with_credentials() {
         let url_str = "https://user:pass@pypi-proxy.fly.dev/basic-auth/simple";
         let log_safe_url = DisplaySafeUrl::parse(url_str).unwrap();
         assert_eq!(
-            &log_safe_url.displayable_with_credentials().to_string(),
+            log_safe_url.displayable_with_credentials().to_string(),
             url_str
         );
     }
@@ -301,19 +420,37 @@ mod tests {
         let url_str = "https://token@example.com/abc/";
         let log_safe_url = DisplaySafeUrl::parse(url_str).unwrap();
         let foo_url = log_safe_url.join("foo").unwrap();
-        assert_eq!(format!("{foo_url}"), "https://****@example.com/abc/foo");
+        assert_eq!(foo_url.to_string(), "https://****@example.com/abc/foo");
     }
 
     #[test]
     fn log_safe_url_ref() {
         let url_str = "https://user:pass@pypi-proxy.fly.dev/basic-auth/simple";
-        let url = Url::parse(url_str).unwrap();
+        let url = DisplaySafeUrl::parse(url_str).unwrap();
         let log_safe_url = DisplaySafeUrl::ref_cast(&url);
         assert_eq!(log_safe_url.username(), "user");
         assert!(log_safe_url.password().is_some_and(|p| p == "pass"));
         assert_eq!(
-            format!("{log_safe_url}"),
+            log_safe_url.to_string(),
             "https://user:****@pypi-proxy.fly.dev/basic-auth/simple"
         );
     }
+
+    #[test]
+    fn parse_url_ambiguous() {
+        for url in &[
+            "https://user/name:password@domain/a/b/c",
+            "https://user\\name:password@domain/a/b/c",
+            "https://user#name:password@domain/a/b/c",
+            "https://user.com/name:password@domain/a/b/c",
+        ] {
+            let err = DisplaySafeUrl::parse(url).unwrap_err();
+            match err {
+                DisplaySafeUrlError::AmbiguousAuthority(redacted) => {
+                    assert!(redacted.starts_with("https:***@domain/a/b/c"));
+                }
+                DisplaySafeUrlError::Url(_) => panic!("expected AmbiguousAuthority error"),
+            }
+        }
+    }
 }
